module(..., package.seeall)
local s = require'scheme'
local iface = require'nsgifaceconf'

local ip = "/usr/sbin/ip"

local function getNetDevices()
  local procNetDev, err = io.open"/proc/net/dev"
  if not procNetDev then
    error(err)
  end
  local s = procNetDev:read'*a'
  local t = {}
  for dev in s:gmatch"\n%s*([^%s]*):" do
    t[#t+1] = dev
    t[dev] = true
  end
  return t
end

local function badIfaceForBridgeGroup(newiface, brif, invtab)
  for iface in pairs(brif) do
    for _, set in pairs(invtab) do
      if set[newiface] and set[iface] then
        return iface
      end
    end
  end
end

local function bridgeVal(val)
  val = (tostring(val)):match("^br(%d+)$") or val
  local n, err = s.integerVal(tonumber(val), 1, 255)
  if n then
    local brname = "br"..tostring(n)
    return brname
  else
    return nil, err
  end
end

local function getBridgeVal(ifaceName)
  return function(val)
    local res, err = bridgeVal(val)
    if res then
      local brIf = C.getOption({"bridge", res}, "brIf")
      if brIf then
        if BI.invalidBridgeGroup then
          local conflict = badIfaceForBridgeGroup(ifaceName, brIf, BI.invalidBridgeGroup)
          if conflict then
            return nil, "other bridge-group  because of "..ifaceName..
            " can't be in the same bridge group with "..conflict
          end
        end
      end    
    end
    return res, err
  end
end

local function applyBridgeGroup(path, bridge, ifaceName)
  local oldBridge = C.getOption(path, "bridge-group")
  if oldBridge ~= bridge then
    if oldBridge then
      os.execute("/sbin/brctl delif "..oldBridge..' '..ifaceName.." >/dev/null 2>&1")
      local brIf = C.getOption({"bridge", oldBridge}, "brIf")
      if brIf then
        brIf[ifaceName] = nil
      end
    end

    local brIf = C.getOption({"bridge", bridge}, "brIf")
    if bridge then
      if brIf then
        if BI.invalidBridgeGroup then
          local res = badIfaceForBridgeGroup(ifaceName, brIf, BI.invalidBridgeGroup)
          if res then
            return false, ifaceName.." can't be in the same bridge group with "..res
          end
        end
        brIf[ifaceName] = true
      else
        C.setOption({"bridge", bridge}, "brIf", {[ifaceName] = true})
      end
      os.execute("/sbin/brctl addif "..bridge..' '..ifaceName.." >/dev/null 2>&1")
    else
      if brIf then brIf[ifaceName] = nil end
    end
    C.setOption(path, "bridge-group", bridge)
  end
  return true
end

ethScheme = iface.interfaceScheme{
  _Z = function(path, sch, conf) 
    local ifaceName = path[#path]
    iface.addIfaceAlias(ifaceName, ifaceName)
    local resB, errB = applyBridgeGroup(
      path, 
      (conf or {})["bridge-group"], 
      ifaceName
    )
    local res, err = iface.interfaceScheme._Z(path, sch, conf)
    return resB and res, table.concat({errB or "", err or ""}, '\n') 
  end,

  ["bridge-group"] = BI.bridge and function(_, path) 
    return s.newScheme{
      _H = "ethBridgegroup",
      _T = "string",
      _V = getBridgeVal(path[#path - 1]),
    }
  end,
}

local vlanScheme = ethScheme{
  _H = "ethVlanN",
  _Z = function(path, sch, conf) 
    local devname = path[#path]
    local ethN, vlanID = devname:match"^(.*)%.(%d+)$"
    if not getNetDevices()[devname] then
      os.execute("/sbin/vconfig add "..ethN.." "..vlanID.." >/dev/null 2>&1")
    end
    return ethScheme._Z(path, sch, conf)
  end,

}

ethScheme.vlan = function(_, path)
  return s.newScheme{
    _H = "ethVlan",
    _Z = function(path, sch, conf)
      local netdevs = getNetDevices()
      local baseIfaceName = path[#path - 1]
      for i = 1, #netdevs do
        local dev = netdevs[i]
        if ( dev:match("^"..baseIfaceName.."%.%d+$") and 
          not (conf or {})[dev]
        ) then
          os.execute("/sbin/vconfig rem "..dev.." >/dev/null 2>&1")
        end
      end
      return true
    end,
    _NAME = vlanScheme,
    _NAME_TYPE = function(val)
      local portName = path[#path - 1]
      val = (tostring(val)):match("^"..portName.."%.(%d+)$") or val
      local n, err = s.integerVal(tonumber(val), 1, 4094)
      if n then
        return portName..'.'..tostring(n)
      else
        return nil, err
      end
    end,
  }
end


--------------------------------------------------------------------------------
-- bridge stuff
--

local par2brctl = {
  ["aging-time"] = "setageing",
  ["forward-time"] = "setfd",
  ["hello-time"] = "sethello",
  ["max-age"] = "setmaxage", 
  ["priority"] = "setbridgeprio",
  ["stp"] = "stp",
}

local function setBridgeParams(devname, conf)
  for k, v in pairs(par2brctl) do
    if conf[k] then
      os.execute("/sbin/brctl "..v.." "..devname.." "..conf[k].." >/dev/null 2>&1")
      print("/sbin/brctl "..v.." "..devname.." "..conf[k].." >/dev/null 2>&1")
    end
  end
end

bridgeNScheme = ethScheme{
  _H = "bridgeN",

  _Z = function(path, sch, conf) 
    local devname = path[#path]
    if not getNetDevices()[devname] then
      os.execute("/sbin/brctl addbr "..devname.." >/dev/null 2>&1")
    end
    local brIf = C.getOption({"bridge", devname}, "brIf") or{}
    for ifaceName in pairs(brIf) do
      os.execute("/sbin/brctl addif "..devname..' '..ifaceName.." >/dev/null 2>&1")
    end
    conf = s.addDefault(path, sch, conf or {}, true) --clone conf with defaults
    setBridgeParams(devname, conf)
    return ethScheme._Z(path, sch, conf)
  end,

  ["bridge-group"] = s.nilType,

  ["aging-time"] = s.integerType{
    _H = "bridgeAgingtime",
    _A = {10, 1000000},
    _D = 300,
  },
  ["forward-time"] = s.integerType{
    _H = "bridgeForwardtime",
    _A = {4, 200},
    _D = 15,
  },
  ["hello-time"] = s.integerType{
    _H = "bridgeHellotime",
    _A = {1, 10},
    _D = 2,
  },
  ["max-age"] = s.integerType{
    _H = "bridgeMaxage",
    _A = {6, 200},
    _D = 20,
  },
  ["priority"] = s.integerType{
    _H = "bridgePriority",
    _A = {0, 65535},
    _D = 32768,
  },
  ["stp"] =  s.newScheme{
    _H = "bridgeStp",
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _D = "on"
  },

  showmacs = s.newScheme{
    _H = "bridgeShowmacs",
    _R = 'user',
    _X = function(path)
      return s.readpopen("/sbin/brctl showmacs "..path[#path].." 2>&1")
    end,
  },

  show = s.newScheme{
    _H = "bridgeShow",
    _R = 'user',
    _X = function(path)
      local dev = path[#path - 1]
      local stp = s.readpopen("/sbin/brctl showstp "..dev.." 2>&1")
      local ifstate = iface.showInterfaceStat(dev)
      return stp..ifstate
    end,
  },
}

bridgeScheme = s.newScheme{
  _H = "bridge",
  _Z = function(path, sch, conf)
    local netdevs = getNetDevices()
    for i = 1, #netdevs do
      local dev = netdevs[i]
      if ( dev:match("^br%d+$") and 
        not (conf or {})[dev]
      ) then
        os.execute(
          ip.." link set "..dev.." down;"..
          "/sbin/brctl delbr "..dev.." >/dev/null 2>&1"
        )
      end
    end
    return true
  end,
  _NAME = bridgeNScheme,
  _NAME_TYPE = bridgeVal,

  show = s.newScheme{
    _H = "bridgeShow",
    _R = 'user',
    _X = function()
      return s.readpopen("/sbin/brctl show 2>&1")
    end,
  },
}

--------------------------------------------------------------------------------
local ethPhyScheme = s.integerType{
  _H = "ethernetSwitchPhy",
  _A = {0,4},
}

local swvlan2nsg = {
  [0] = 4,
  [1] = 3,
  [2] = 2,
  [3] = 1,
  [4] = 0,
}

local swport2nsg = {
  ['0'] = 4,
  ['1'] = 3,
  ['2'] = 2,
  ['3'] = 1,
  ['4'] = 0,
}

ethernetSwitchScheme = s.newScheme{
  _H = "ethernetSwitch",
  _Z = function(path, sch, conf)
    conf = s.addDefault(path, sch, conf or {}, true) --clone conf with defaults
    local matrix = {
      [0] = {},
      [1] = {},
      [2] = {},
      [3] = {},
      [4] = {},
    }
    for k, v in pairs(conf) do
      local phy = k:match"^phy(%d)$"
      if phy then
        matrix[v][#matrix[v]+1] = phy
      end
    end
    for i = 0, 4 do
      local s = table.concat(matrix[i]):gsub('%d', swport2nsg)
      if #s > 0 then s = s..'c' end
      local com = '/sbin/admswconfig eth'..i..' "'..s..'"'
      os.execute(com.." >/dev/null 2>&1")
    end
    return true
  end,
  phy0 = ethPhyScheme{_D = 0},
  phy1 = ethPhyScheme{_D = 1},
  phy2 = ethPhyScheme{_D = 2},
  phy3 = ethPhyScheme{_D = 3},
  phy4 = ethPhyScheme{_D = 4},
  show = s.newScheme{
    _H = "ethernetSwitchShow",
    _X = function(path)
      local info = s.readpopen("/sbin/admswconfig")
      local matrix = {}
      for vlan, p4, p3, p2, p1, p0, cpu in 
      info:gmatch"\n(%d)\t(1?)\t(1?)\t(1?)\t(1?)\t(1?)\t(1?)" do
        matrix[swvlan2nsg[tonumber(vlan)]] = {
          [0] = p0 == '1' and true, 
          [1] = p1 == '1' and true, 
          [2] = p2 == '1' and true, 
          [3] = p3 == '1' and true, 
          [4] = p4 == '1' and true,
        }
      end
      local out = {'\n      phy0 phy1 phy2 phy3 phy4'}
      for i = 0, 4 do
        local vlan = {}
        vlan[#vlan+1] = "eth"..i
        for j = 0, 4 do
          vlan[#vlan+1] = matrix[i][j] and 'X' or ' '
        end
        out[#out+1] = table.concat(vlan, '    ')
      end

      return table.concat(out, '\n')..'\n'
    end,
  },
}
