module(..., package.seeall)
local s = require'scheme'

local ip = "/usr/sbin/ip"

local ifaceAliases = {}
local ifaceAliasesHooks = {}
local function callHooks()
  for _, f in pairs(ifaceAliasesHooks) do
    f()
  end
end
function setAliasHook(k, f)
  ifaceAliasesHooks[k] = f
end
function addIfaceAlias(alias, name)
  if ifaceAliases[alias] ~= name then
    ifaceAliases[alias] = name
    if alias ~= name then
      callHooks()
    end
  end
end
function getIfaceName(alias)
  return ifaceAliases[alias] or alias
end
function ifaceExist(name)
  for k, v in pairs(ifaceAliases) do
    if v == name then
      return k
    end
  end
  return false
end

local function setIfAddress(devName, ifAddress)
  if ifAddress and ifAddress.prefix then
    local peer, broadcast, anycast = '', '', ''
    if ifAddress.peer then
      peer = "peer "..ifAddress.peer
    end
    if ifAddress.broadcast then
      broadcast = "broadcast "..ifAddress.broadcast
    end
    if ifAddress.anycast then
      anycast = "anycast "..ifAddress.anycast
    end
    local com = ("%s addr add dev %s %s %s %s %s 2>&1"):format(ip, devName,
    ifAddress.prefix, peer, broadcast, anycast)
    local res = s.readpopen(com)
    if not res:match'^%s*$' then
      C.log_warn(com..'\n'..res)
    else
      C.log_info(com)
    end
  end
end

local function setIfAddresses(devName, ifAddress)
  if ifAddress then
    setIfAddress(devName, ifAddress)
    for _, v in ipairs(ifAddress) do
      setIfAddress(devName, v)
    end
  end
end

local devIfAddressesConfigs = {}
local applyIfAddresses =
function(dev)
  if dev and devIfAddressesConfigs[dev] then
    local cfg = devIfAddressesConfigs[dev]
    if cfg.conf and cfg.conf.configurable == "dhcp" then
      local Cmd = {"udhcpc", "-f", "-i", dev, } -- foreground
      os.execute(ip..' addr flush dev '..dev.." >/dev/null 2>&1")
      C.restartDaemon(cfg.path, Cmd)
      C.fakeLastComTimestamp() -- user may lose connection when ip addr changed
    else
      C.restartDaemon(cfg.path)  -- kill "daemon" if not needed
      if cfg.conf and cfg.conf.configurable == "yes" then
        os.execute(ip..' addr flush dev '..dev.." >/dev/null 2>&1")
        setIfAddresses(dev, cfg.conf)
        C.fakeLastComTimestamp() -- user may lose connection when ip addr changed
      end
    end
  end
end

local resetIfAddressesNeteventKey = {}
local resetIfAddressesNeteventCallBack =
function(msg)
  if (msg.type == 'newlink' and not msg.ifup) then
    applyIfAddresses(msg.ifname)
  end
end

function resetIfAddresses(path, sch, conf, prevConf)
  conf = s.addDefault(path, sch, conf, true)
  local devPath = s.deepcopy(path)
  devPath[#devPath ] = "ifName"
  local dev = C.getConfOrDefault(devPath)  or getIfaceName(path[#path - 1])

  --save current config to global table
  devIfAddressesConfigs[dev] = {
    conf = conf,
    path = s.deepcopy(path),
  }
  C.updateNeteventCallback( -- try set on dev creation
    resetIfAddressesNeteventKey,
    resetIfAddressesNeteventCallBack
  )
  applyIfAddresses(dev) -- try set now
  return true
end

local devIpLinkConfigs = {}

function applyIpLink(dev)
  local conf = devIpLinkConfigs[dev] and devIpLinkConfigs[dev].conf
  if conf then
    local com = {
      ip, 'link', 'set', dev,
    }
    if conf.arp then
      com[#com+1] = 'arp'
      com[#com+1] = conf.arp
    end
    if conf.multicast then
      com[#com+1] = 'multicast'
      com[#com+1] = conf.multicast
    end
    if conf.dynamic then
      com[#com+1] = 'dynamic'
      com[#com+1] = conf.dynamic
    end
    if conf.qlen then
      com[#com+1] = 'txqueuelen'
      com[#com+1] = conf.qlen
    end
    if conf.mtu then
      com[#com+1] = 'mtu'
      com[#com+1] = conf.mtu
    end
    if conf['mac-address'] --and 
    --conf['mac-address'] ~= (prevConf or {})['mac-address'] 
    then
      com[#com+1] = 'address'
      com[#com+1] = conf['mac-address']
      os.execute(ip.." link set "..dev.." down")
      com[#com+1] = devIpLinkConfigs[dev]["adm-state"]
    end
    com[#com+1] = "2>&1"

    if #com>5 then --may be nothing to do ?
      local res = s.readpopen(table.concat(com,' '))
      if not res:match'^%s*$' then
        C.log_warn(table.concat(com,' ')..'\n'..res)
      else
        C.log_info(table.concat(com,' '))
      end
    end
  end
  return true
end

local resetIpLinkNeteventKey = {}
local resetIpLinkNeteventCallBack =
function(msg)
  if (msg.type == 'newlink' and not msg.ifup) then
    applyIpLink(msg.ifname)
  end
end

function ipLinkSet(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
    local devPath = s.deepcopy(path)
    devPath[#devPath] = "ifName"
    local dev =  C.getConfOrDefault(devPath)  or getIfaceName(path[#path - 1])
    --save current config to global table
    devIpLinkConfigs[dev] = {
      conf = conf,
      ["adm-state"] = C.getConfOrDefault(s.clonePath(path, -1, "adm-state"))
    }
    C.updateNeteventCallback( -- try set on dev creation
      resetIpLinkNeteventKey,
      resetIpLinkNeteventCallBack
    )
    applyIpLink(dev) -- try set now
  end
  return true
end

local ifAddressScheme = s.newScheme{
  _H = "interface address",
  _N = function(sch, conf)
    if conf and not conf.prefix then
      return {nil, peer = true, broadcast = true, anycast = true}
    end
    return {}
  end, 
  prefix = s.ipPrefixType{_H = "ifAddressPrefix", _P = 100, _DEP = 1},
  peer = s.ipAddressType{_H = "ifAddressPeer", _P = 90},
  broadcast = s.ipAddressType{_H = "ifAddressBroadcast", _P = 80},
  anycast = s.ipAddressType{_H = "ifAddressAnycast", _P = 70},
}

ifAddressListScheme = ifAddressScheme{
  _H = "ifAddress",
  _Z = resetIfAddresses,
  configurable =  s.newScheme{
    _T = "choice",
    _A = {
      {"yes", ""},
      {"no", ""},
      {"dhcp", ""},
    },
    _V = s.choiceVal,
    _H = "ifAddressEnable",
    _D = "yes",
  },
  _LIST = ifAddressScheme{
    _H = "ifAddressSecondary",
  }
}

linkScheme = s.newScheme{
  _H = "ifLink",
  _Z = ipLinkSet,
  arp =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkArp",
  },
  multicast =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkMulticast",
  },
  dynamic =  s.newScheme{
    _T = "choice",
    _A = {"on", "off"},
    _V = s.choiceVal,
    _H = "ifLinkDynamic",
  },
  qlen =  s.newScheme{
    _T = "number",
    _A = {0, 2147483647},
    _V = s.integerVal,
    _H = "ifLinkTxlen",
  },
  mtu =  s.newScheme{
    _T = "number",
    _A = {0, 9000},
    _V = s.integerVal,
    _H = "ifLinkMtu",
  },
  ['mac-address'] = s.newScheme{
    _T = 'string',
    _V = s.macaddressVal,
    _H = "ifLinkMacaddress",
  },
}

function showInterfaceStat(iface)
  local addr = s.readpopen(ip..' addr show dev '..iface.." 2>&1")
  local link = s.readpopen(ip..' -s -s link show dev '..iface.." 2>&1")
  link = link:gsub("^.-\n.-\n(.*)$", "%1")
  return addr..link
end

function interfaceMainApply(path, sch, conf)
  local devName = C.getConfOrDefault(s.clonePath(path, nil, "ifName"))  
  or getIfaceName(path[#path])
  local adm = C.getConfOrDefault(s.clonePath(path, nil, "adm-state"))
  -- (conf or {})["adm-state"] or sch["adm-state"]._D or "down"
  local com = {ip, 'link', 'set', devName, adm, "2>&1"}
  os.execute(table.concat(com,' '))
  local res = s.readpopen(table.concat(com,' '))
  if not res:match'^%s*$' then
    C.log_warn(table.concat(com,' ')..'\n'..res)
  else
    C.log_info(table.concat(com,' '))
  end
  return true
end

interfaceScheme = s.newScheme{
  _H = "if1",
  _Z = function(path, sch, conf)
    return interfaceMainApply(path, sch, conf)
  end,
  ["adm-state"] =  s.newScheme{
    _T = "choice",
    _A = {"up", "down"},
    _V = s.choiceVal,
    _H = "ifLinkAdm",
    _D = "up"
  },
  ifAddress = ifAddressListScheme,
  link = linkScheme,
  show = s.newScheme{
    _H = "ifShow",
    _X = function(path)
      return showInterfaceStat(getIfaceName(path[#path-1]))
    end,
    _R = 'user',
  },
}

local LO_PREFIX = "127.0.0.1/8"
local loInterfaceScheme = s.newScheme{
  _H = "ifLo",
  _Z = function() return true end,
  ifAddress = s.newScheme{
    _H = "ifAddress",
    _Z = resetIfAddresses,
    prefix = s.newScheme{
      _T = "string",
      _V = function(val)
        return val == LO_PREFIX and LO_PREFIX or
        nil, LO_PREFIX
      end,
      _H = "ifAddressPrefix",
      _D = LO_PREFIX,
    },
    configurable =  s.newScheme{
      _T = "choice",
      _H = "ifAddressEnable",
      _A = {"yes", "no", "dhcp"},
      _V = s.choiceVal,
      _D = "no",
    },
    _LIST = ifAddressScheme{
      _H = "ifAddressSecondary",

    },
  },
  show = s.newScheme{
    _H = "ifShow",
    _X = function(path)
      return showInterfaceStat(path[#path-1])
    end,
    _R = 'user',
  },
}

interfaceListScheme = s.newScheme{
  _H = "if0",
  _Z = function() return true end,
  lo = loInterfaceScheme,
  _NAME = interfaceScheme{
    _Z = function(path, sch, conf) 
      local alias = ifaceExist(path[#path])
      if alias then
        C.setConfigValue(path, nil)
        return false, "There is node "..alias.." used this ip interface" 
      end
      return interfaceMainApply(path, sch, conf)
    end,
  },
  _NAME_TYPE = function(val)
    local res, err =  s.stringVal(val, 16)
    if not res then
      return res, err
    end
    local alias = ifaceExist(val)
    if alias then
      return false, "interface name except lo, "..alias..'...'
    end
    return val
  end,
}

